#include "MPU9250.h"
#include "timers.h"
#include "fingers.h"
#include "serial_com.h"

MPU9250 mpu;
#define X 0
#define Y 1
#define Z 2
#define MIN_X 3
#define MIN_Y 4
#define MIN_Z 5

#define STATE_WAIT_INPUT 0
#define STATE_TRANSLATE 1
#define STATE_CORRECT 2
#define STATE_UNDEFINED 3
#define STATE_RECALIBRATE 4
#define STATE_STANDBY 5
#define STATE_CONTINUE 6

float a_x = 0, a_y = 0, a_z = 0;
float biasZ, biasX, biasY;
int up, prev_up;
int prev_gesture_state, gesture_state;
float ang_accX, ang_accY, ang_accZ;

//  shared var
volatile byte state, prev_state;
volatile byte start_calibration, count;

int avg0, avg1, avg2, avg3, avg4;
int sign, counter, counter0, counter1;
bool bent0, bent1, bent2, bent3, bent4;

//  handle interrupts every 1ms
ISR(TIMER0_COMPA_vect) {
  if (count) {
    counter++;
    if (counter == 4000) {
      state = STATE_UNDEFINED;
      count = 0;
      counter = 0;
    }
  } else {
    if (counter)
      counter = 0;
  }
}

//  handle interrupts every 1ms
ISR(TIMER1_COMPA_vect) {
  // calibrate for aprox 4s
  if (start_calibration) {
    counter1++;
    if (counter1 == 4000) {
      if (prev_state == STATE_UNDEFINED) {
        state = STATE_TRANSLATE;
        bluetooth_send("Try again!\n");
      } else {
        state = STATE_WAIT_INPUT;
      }
      start_calibration = 0;
      counter1 = 0;
    }
  }
}

void get_bent()
{
  bent0 = (avg0 >= 700) ? 1 : 0;
  bent2 = (avg2 >= 650) ? 1 : 0;
  bent3 = (avg3 >= 700) ? 1 : 0;
  bent4 = (avg4 >= 700) ? 1 : 0;
}

void get_angles() {
  float g_x = mpu.getAccX();
  float g_y = mpu.getAccY();
  float g_z = mpu.getAccZ();

  if(g_x > 1.0) g_x = 1.0;
  if(g_x < -1.0) g_x = -1.0;

  a_x = (asin(g_x)) * 57.296;

  if(g_y > 1.0) g_y = 1.0;
  if(g_y < -1.0) g_y = -1.0;

  a_y = (asin(g_y)) * 57.296;

  if(g_z > 1.0) g_z = 1.0;
  if(g_z < -1.0) g_z = -1.0;

  a_z = (asin(g_z)) * 57.296;
}

void get_hand_orientation() {
  get_angles();
  ang_accZ = mpu.getGyroZ() - biasZ;
  ang_accX = mpu.getGyroX() - biasX;
  ang_accY = mpu.getGyroX() - biasY;
  int copy_up = up;
  float rotation;

  if(abs(a_x) < 45){
      if(abs(a_y) < 45){
        up = (a_z >= 0) ? Z : MIN_Z;
      } else{
        up = (a_y > 0) ? Y : MIN_Y;
      }
  } else{
    up = (a_x > 0) ? X : MIN_X;
  }

  if (up != copy_up) {
    prev_up = copy_up;
  }
}

void record_gestures()
{
  // get flex sensors values
  avg0 += analogRead(LITTLE);
  avg0 >>= 1;

  // avg1 += analogRead(RING);
  // avg1 >>= 1;

  avg2 += analogRead(MIDDLE);
  avg2 >>= 1;

  avg3 += analogRead(INDEX);
  avg3 >>= 1;

  avg4 += analogRead(THUMB);
  avg4 >>= 1;

  print_data();
  get_bent();

  //  get gyro and acc data
  if (mpu.update())
    get_hand_orientation();
}

void setup() {
  state = STATE_WAIT_INPUT;
  gesture_state = 0;
  counter0 = counter = 0;
  count = 0;
  start_calibration = 0;

  Serial.begin(115200);
  Wire.begin();
  delay(2000);

  //  set serial connection 
  set_bluetooth_connection();

  // start
  int message = bluetooth_recv();
  while (message != START)
    message = bluetooth_recv();

  if (!mpu.setup(0x68)) {
    while (1) {
      bluetooth_send("MPU connection failed.\n");
      delay(5000);
    }
  }

  // calibrate acc & gyro
  calibrate_device();

  // set pins for flex sensors 
  fingers_setup();

  //  set timers
  cli();
  timer0_setup();
  timer1_setup();
  sei();
}

void loop() {
  record_gestures();

  switch(state) {

    case STATE_WAIT_INPUT:
      bluetooth_send("You have to sign...");
      sign = bluetooth_recv();

      if (sign == END) {
        bluetooth_send("\nPress START to practice.\n");
        state = STATE_STANDBY;
        break;
      }

      state = STATE_TRANSLATE;
      break;
    case STATE_TRANSLATE:
      if (sign == HELLO) {
          bool all_straight = !(bent0 | bent2 | bent3 | bent4);

          switch(gesture_state) {
            case 0:
              count = 1;
              if (up == Y && all_straight) {
                gesture_state++;
                count = 0;
              }

              break;
            case 1:
              count = 1;
              if (ang_accZ > 50 && all_straight) {
                gesture_state++;
                count = 0;
              }
                          
              break;
            case 2:
              count = 1;
              if (up == MIN_X && all_straight) {
                gesture_state++;
                count = 0;
              }
              break;
            case 3:
              count = 1;
              if (ang_accZ < -50 && all_straight) {
                gesture_state++;
                count = 0;
              }
              break;
            case 4:
              count = 1;
              if (up == Y && all_straight) {
                gesture_state++;
                count = 0;
              }
              break;
            case 5:
              count = 0;
              state = STATE_CORRECT;
              gesture_state = 0;
              break;
            default:
              break;
          }
        break;
      }

      if (sign == YES) {
        bool all_bent = bent0 & bent2 & bent3 &bent4;

        switch(gesture_state) {
          case 0:
            count = 1;
            if (up == Y && all_bent) {
              gesture_state++;
              count = 0;
            }
              
            break;
          case 1:
            count = 1;
            if (ang_accX < -50 && all_bent) {
              gesture_state++;
              count = 0;
            }
            
            break;
          case 2:
            count = 1;
            if (up == Z && all_bent) {
              gesture_state++;
              count = 0;
            }
            break;
          case 3:
            count = 0;
            state = STATE_CORRECT;
            gesture_state = 0;
          default:
            break;
        }
        break;
      }

      if (sign == THANK_YOU) {
          bool all_straight = !(bent0 | bent2 | bent3 | bent4);

          switch(gesture_state) {
          case 0:
            count = 1;
            if (up == Y && all_straight) {
              gesture_state++;
              count = 0;
            }
              
            break;
          case 1:
            count = 1;
            if (ang_accX < -50 && all_straight) {
              gesture_state++;
              count = 0;
            }
            
            break;
          case 2:
            count = 1;
            if (up == MIN_Z  && all_straight) {
              gesture_state++;
              count = 0;
            }
            break;
          case 3:
            count = 0;
            state = STATE_CORRECT;
            gesture_state = 0;
            break;
          default:
            break;
        }
        break;
      }

      if (sign == ILY) {
          bool straight_034 = !(bent0 | bent3 | bent4);
          bool bent_12 = bent2;

          switch(gesture_state) {
            case 0:
              count = 1;
              if (up == Y && straight_034 && bent_12) {
                gesture_state++;
                count = 0;
              }
              break;
            case 1:
              count = 1;
              if (ang_accZ > 50 && straight_034 && bent_12) {
                gesture_state++;
                count = 0;
              }     
              break;
            case 2:
              count = 1;
              if (up == MIN_X && straight_034 && bent_12) {
                gesture_state++;
                count = 0;
              }
              break;
            case 3:
              count = 1;
              if (ang_accZ < -50 && straight_034 && bent_12) {
                gesture_state++;
                count = 0;
              }
              break;
            case 4:
              count = 1;
              if (up == Y && straight_034 && bent_12) {
                gesture_state++;
                count = 0;
              }
              break;
            case 5:
              count = 0;
              state = STATE_CORRECT;
              gesture_state = 0;
              break;
            default:
              break;
          }
        break;
      }

      if (sign == YOU) {
        if (up == MIN_X && !bent3 && (bent0 && bent2 && bent4)) {
          state = STATE_CORRECT;
          gesture_state = 0;
        }
        break;
      }

      if (sign == WELCOME) {
         bool all_straight = !(bent0 | bent2 | bent3 | bent4);
          switch(gesture_state) {
          case 0:
            count = 1; 
            if (up == Z && all_straight) {
                gesture_state++;
                count = 0;
              }
            break;
          case 1:
            count = 1;
            if (ang_accY > 50 && all_straight) {
                gesture_state++;
                count = 0;
              }
            
            break;
          case 2:
            count = 1;
            if (up == MIN_Z && all_straight) {
                gesture_state++;
                count = 0;
              }
            break;
          case 3:
            count = 0;
            state = STATE_CORRECT;
            gesture_state = 0;
            break;
          default:
            break;
        }
        break;
      }
      break;
    
    //  the user signed correctly 
    case STATE_CORRECT:
        bluetooth_send("You used the RIGHT signing!\n");
  
        //  delay to make sure sensors recalibrate
        prev_state = state;
        state = STATE_RECALIBRATE;
        bluetooth_send("Wait for device to recalibrate...\n");
        break;

    //  the user signed incorrectly
    case STATE_UNDEFINED:
        bluetooth_send("WRONG SIGN!\n");
        bluetooth_send("Wait for device to recalibrate...\n");

        //  delay to make sure sensors recalibrate
        prev_state = state;
        state = STATE_RECALIBRATE;

        gesture_state = 0;
        break;

    //  dummy state for delay
    case STATE_RECALIBRATE:
        //  start recalibration counter
        start_calibration = 1;
        break;

    case STATE_STANDBY:
      int message = bluetooth_recv();

      if (message == START) {
        state = STATE_WAIT_INPUT;
      }
      break;
    default:
      state = STATE_STANDBY;
      break;
  }
}

void get_calibration() {
  biasX = mpu.getGyroBiasX() / (float)MPU9250::CALIB_GYRO_SENSITIVITY;
  biasY = mpu.getGyroBiasY() / (float)MPU9250::CALIB_GYRO_SENSITIVITY;
  biasZ = mpu.getGyroBiasZ() / (float)MPU9250::CALIB_GYRO_SENSITIVITY;
}

void calibrate_device()
{
  bluetooth_send("Calibration will start in 5 sec.\n");
  bluetooth_send("Leave your palm still on a flat plane.\n");
  mpu.verbose(true);
  delay(5000);
  mpu.calibrateAccelGyro();
  get_calibration();
  mpu.verbose(false);
}

void print_data() {
  Serial.print(avg4);
  Serial.print(" ");
  Serial.print(avg3);
  Serial.print(" ");
  Serial.print(avg2);
  Serial.print(" ");
  Serial.println(avg0);
}